#include "string.h"
#include "talloc.h"
#include "dynamic.c"

/* Iterator over the pdf function */
/* Keeping it in the header allows to inline pdf calls if necessary */
double* EF(iter_pdf)(family* ef, unsigned int n, double* in, param* lambda, double* out) {
  unsigned int i;
  for(i=0; i<n; i++) {
    out[i] = EF(pdf)(ef, &in[i], lambda);
  }
  return out;
}

void EF(theta2eta)(family* ef, param* theta, param* eta) {
  EF(gradF)(ef, theta, eta);
}
void EF(eta2theta)(family* ef, param* eta, param* theta) {
  EF(gradG)(ef, eta, theta);
}

param* EF(mle)(family* ef, unsigned int n, double* data, char* mask, param* eta) {
  CHECK_EF(ef, eta);
  CHECK_TYPE(eta, EXPECTATION);

  unsigned int count = 0;

  for (unsigned int i=0; i<n; i++) {
    if (mask[i]) {
      param* t = EF(new_param)(ef, EXPECTATION);
      EF(t)(ef, &data[i * ef->dim], t);
      EF(add)(ef, eta, t, eta);
      count++;
    }
  }
  EF(scale)(ef, eta, 1./count, eta);

  return eta;
}

void* EF(new_param)(family* ef, enum type type) {
  param* p = NULL;
  switch (type) {
  case SOURCE:
    p = malloc(sizeof(struct source));
    EF(dynamic_alloc)(EF(source_descr), EF(source_len), ef->dim, p);
    break;
  case NATURAL:
    p = malloc(sizeof(struct natural));
    EF(dynamic_alloc)(EF(natural_descr), EF(natural_len), ef->dim, p);
    break;
  case EXPECTATION:
    p = malloc(sizeof(struct expectation));
    EF(dynamic_alloc)(EF(expectation_descr), EF(expectation_len), ef->dim, p);
    break;
  }
  
  p->type = type;
  p->ef   = (void*)ef; /* type param only accepts real family, not
                          XXX_family (in which the family in the
                          prototype is translated */

  return p;
}

void EF(delete_param)(family* ef, param* p) {
  switch (p->type) {
  case SOURCE:
    EF(dynamic_free)(EF(source_descr), EF(source_len), p);
    break;
  case NATURAL:
    EF(dynamic_free)(EF(natural_descr), EF(natural_len), p);
    break;
  case EXPECTATION:
    EF(dynamic_free)(EF(expectation_descr), EF(expectation_len), p);
    break;
  }

  free(p);
}

void EF(copy_param)(family* ef, param* src, param* dest) {
  CHECK_EF(ef, src);
  CHECK_EF(ef, dest);
  assert(src->type == dest->type);

  switch (src->type) {
  case SOURCE:
    EF(dynamic_copy)(EF(source_descr), EF(source_len), src, dest);
    break;
  case NATURAL:
    EF(dynamic_copy)(EF(natural_descr), EF(natural_len), src, dest);
    break;
  case EXPECTATION:
    EF(dynamic_copy)(EF(expectation_descr), EF(expectation_len), src, dest);
    break;
  }
}

param* EF(add)(family* ef, param* p, param* q, param* dest) {
  CHECK_EF(ef, p);
  CHECK_EF(ef, q);
  assert(p->type == q->type);

  switch (p->type) {
  case SOURCE:
    EF(dynamic_add)(EF(source_descr), EF(source_len), p, q, dest);
    break;
  case NATURAL:
    EF(dynamic_add)(EF(natural_descr), EF(natural_len), p, q, dest);
    break;
  case EXPECTATION:
    EF(dynamic_add)(EF(expectation_descr), EF(expectation_len), p, q, dest);
    break;
  }

  return dest;
}

double EF(scalar)(family* ef, param* p, param* q) {
  CHECK_EF(ef, p);
  CHECK_EF(ef, q);

  return EF(dynamic_scalar)(EF(natural_descr), EF(natural_len), p, q);
}

param* EF(minus)(family* ef, param* p, param* q, param* dest) {
  CHECK_EF(ef, p);
  CHECK_EF(ef, q);
  assert(p->type == q->type);

  switch (p->type) {
  case SOURCE:
    EF(dynamic_minus)(EF(source_descr), EF(source_len), p, q, dest);
    break;
  case NATURAL:
    EF(dynamic_minus)(EF(natural_descr), EF(natural_len), p, q, dest);
    break;
  case EXPECTATION:
    EF(dynamic_minus)(EF(expectation_descr), EF(expectation_len), p, q, dest);
    break;
  }

  return dest;
}

param* EF(scale)(family* ef, param* p, double a, param* dest) {
  CHECK_EF(ef, p);

  switch (p->type) {
  case SOURCE:
    EF(dynamic_scale)(EF(source_descr), EF(source_len), p, a, dest);
    break;
  case NATURAL:
    EF(dynamic_scale)(EF(natural_descr), EF(natural_len), p, a, dest);
    break;
  case EXPECTATION:
    EF(dynamic_scale)(EF(expectation_descr), EF(expectation_len), p, a, dest);
    break;
  }

  return dest;
}

param* EF(zero)(family* ef, param* p) {
  CHECK_EF(ef, p);

  switch (p->type) {
  case SOURCE:
    EF(dynamic_zero)(EF(source_descr), EF(source_len), p);
    break;
  case NATURAL:
    EF(dynamic_zero)(EF(natural_descr), EF(natural_len), p);
    break;
  case EXPECTATION:
    EF(dynamic_zero)(EF(expectation_descr), EF(expectation_len), p);
    break;
  }

  return p;
}

param* EF(as_source)(family* ef, param* in) {
  source* lambda = EF(new_param)(ef, SOURCE);

  switch (in->type) {
  case SOURCE:
    ef->copy_param(ef, in, (param*)lambda);
    break;
  case NATURAL:
    ef->theta2lambda(ef, in, (param*)lambda);
    break;
  case EXPECTATION:
    ef->eta2lambda(ef, in, (param*)lambda);
    break;
  default:
    assert(0);
  }

  return (param*)lambda;
}

param* EF(as_natural)(family* ef, param* in) {
  natural* theta = EF(new_param)(ef, NATURAL);

  switch (in->type) {
  case SOURCE:
    ef->lambda2theta(ef, in, (param*)theta);
    break;
  case NATURAL:
    ef->copy_param(ef, in, (param*)theta);
    break;
  case EXPECTATION:
    ef->eta2theta(ef, in, (param*)theta);
    break;
  default:
    assert(0);
  }

  return (param*)theta;
}

param* EF(as_expectation)(family* ef, param* in) {
  expectation* eta = EF(new_param)(ef, EXPECTATION);

  switch (in->type) {
  case SOURCE:
    ef->lambda2eta(ef, in, (param*)eta);
    break;
  case NATURAL:
    ef->theta2eta(ef, in, (param*)eta);
    break;
  case EXPECTATION:
    ef->copy_param(ef, in, (param*)eta);
    break;
  default:
    assert(0);
  }

  return (param*)eta;
}

#ifndef CONSTRUCTOR
family* NAME(void) {
  family* ef = talloc(NULL, family);

  ef->name           = __S(NAME);
  ef->dim            = 1;
  ef->pdf            = &EF(pdf           );
  ef->iter_pdf       = &EF(iter_pdf      );
  ef->lambda2theta   = &EF(lambda2theta  );
  ef->theta2lambda   = &EF(theta2lambda  );
  ef->lambda2eta     = &EF(lambda2eta    );
  ef->eta2lambda     = &EF(eta2lambda    );
  ef->theta2eta      = &EF(theta2eta     );
  ef->eta2theta      = &EF(eta2theta     );
  ef->fprint         = &EF(fprint        );
  ef->F              = &EF(F             );
  ef->gradF          = &EF(gradF         );
  ef->G              = &EF(G             );
  ef->gradG          = &EF(gradG         );
  ef->mle            = &EF(mle           );
  ef->new_param      = &EF(new_param     );
  ef->delete_param   = &EF(delete_param  );
  ef->copy_param     = &EF(copy_param    );
  ef->t              = &EF(t             );
  ef->add            = &EF(add           );
  ef->minus          = &EF(minus         );
  ef->scale          = &EF(scale         );
  ef->scalar         = &EF(scalar        );
  ef->zero           = &EF(zero          );
  ef->rand           = &EF(rand          );
  ef->as_source      = &EF(as_source     );
  ef->as_natural     = &EF(as_natural    );
  ef->as_expectation = &EF(as_expectation);

  return ef;
}
#endif


